# Copyright 1999-2014. Parallels IP Holdings GmbH. All Rights Reserved.
import os
import stat

def copy_perms(src, dst):
    """ Copy owner, group and permission bits from src to dst. Links are followed. """
    src_stat = os.stat(src)
    mode = stat.S_IMODE(src_stat.st_mode)
    uid, gid = src_stat.st_uid, src_stat.st_gid
    os.chown(dst, uid, gid)
    os.chmod(dst, mode)

class ReplacementFile:
    """ Creates temporary file that will replace a given one upon exit from this
        context manager. File owner, group and permissions will be preserved.
    """
    def __init__(self, base_filename, mode=0600, owner_id=-1, group_id=-1, temp_filename=None):
        """ Specifies a file to be replaced and default access rights if it doesn't exist. 
            temp_filename if specified is used as a path to temporary file, this should be 
            on the same filesystem as base one.
        """
        self._base_filename = base_filename
        self._mode = mode
        self._uid = owner_id
        self._gid = group_id
        if temp_filename:
            self._mirror_filename = temp_filename
        else:
            self._mirror_filename = self._base_filename + os.extsep + 'tmp'
        self._mirror = None

    def __enter__(self):
        return self._create_mirror_file()

    def __exit__(self, type, value, traceback):
        self._mirror.close()
        self._replace_with_mirror_file()

    def _create_mirror_file(self):
        """ Returns an opened file object for temporary file, which is a 'mirror' for 
            a given one. This original file may be replaced by a 'mirror' later.
        """
        self._mirror = open(self._mirror_filename, 'w')
        try:
            # Not very secure way, but an easy one
            if os.path.exists(self._base_filename):
                copy_perms(self._base_filename, self._mirror_filename)
            else:
                os.chown(self._mirror_filename, self._uid, self._gid)
                os.chmod(self._mirror_filename, self._mode)
        except:
            self._mirror.close()
            os.unlink(self._mirror_filename)
            raise
        return self._mirror

    def _replace_with_mirror_file(self):
        """ Replaces a given file with a previously created 'mirror' file. """
        os.rename(self._mirror_filename, self._base_filename)

if __name__ == '__main__':
    import doctest, sys
    os.mkdir('testdir')
    sys.exit( doctest.testmod().failed )

# vim: ts=4 sts=4 sw=4 et :
